this.name = "Orbits";
this.author = "Ebi";
this.description = "Adds planets on concentric orbits around the sun."
this.version = "1.2";
this.copyright = "This work is hereby placed in the public domain.";

function pow(x, y)
{
    return Math.exp(y * Math.log(x));
}

// Maximum number of planets to be added (See planetinfo.plist as well).
this.MaxPlanets = 7;

// Days per year of the Main planet
this.Yr = 300;

// Get weights by linearily interpolating the planet radii as defined
// by [X0,Y0] and [X1,Y1]. Those weights are used as multiplier when
// calculating orbital periods. If X0 or X1 are 0 or undefined "orbits"
// will select the minimum or maximum from the planets in the
// system.planets list.
this.X0 = 3000 * 10;		// planetinfo.plist maximum in game units
this.X1 = 12000 * 10;		// planetinfo.plist maximum in game units
this.Y0 = pow(2, 3.0 / 4.0);	// Compensate Keplers law and take square root
this.Y1 = pow(3.5, 3.0 / 4.0);  // Thus, scale radius by values in [2,3.5]
// System spread factor
this.SpreadFactor = 4.5;

// Add a random offset in the range [0, t*RandomOffsetFactor] to the
// orbital period t of a planet.
this.RandomOffsetFactor = 0.1;
// Log to logfile (1) and/or console (2)
this.Verbose = 0; //1|4;
// Oh man, it's too tricky to get the OXP initialization order right
// for Oolite. It may be random, therefore I offer the farsun
// functionality here as well. Shifts the sun and increases (or
// decreases if less than 1) its distance to the main planet by this
// factor. Thus, the value 1 will not show any effect.
this.FarsunMultiplier = 1; // 6;

this.Log = function(dest, msg)
{
    if (dest & 1)
	log (msg);
    if (dest & 2)
	player.commsMessage(msg);
}

this.startUp = this.reset = function()
{
    this.systemDone = false;
    this.Log(this.Verbose & 1, "Initialising OXP " + this.name);
}

this.shipWillLaunchFromStation  =  function()
{
    if (!this.systemDone)
	this.createOrbits();
}

this.shipLaunchedFromStation = function()
{
    if (!this.systemDone)
	this.shipExitedWitchspace();
}

this.shipWillExitWitchspace = function()
{
    this.createOrbits();
}

this.shipExitedWitchspace = function()
{
    this.systemDone = true;
    if ((this.Verbose & (2|4)) == 0 || this.innerPlanets == 0)
	return;
    var msg =
	system.name + ": " +
	this.innerPlanets + "+" +
	this.outerPlanets + " planets";
    if (this.info != undefined)
	msg = msg + "[" + this.info + "]";
    this.Log(2, msg);
}

this.isMoon = function(p)
{
    // 10 * Radius from planetinfo.plist
    return !p.isMainPlanet && p.radius < 1000 * 10;
}

function mkPlanetIndex()
{
    var v = system.planets;
    var idx = new Array (v.length);
    var i, j, t;
    for (i = 0; i < v.length; ++i)
	idx[i] = i;
    for (i = 0; i < v.length - 1; ++i)
	for (j = i + 1; j < v.length; ++j)
	    if (v[idx[i]].ID > v[idx[j]].ID)
		t = idx[i], idx[i] = idx[j], idx[j] = t;
    return idx;
}

this.planetsMinMax = function(x)
{
    var needX0 = this.X0 == undefined || this.X0 == 0;
    var needX1 = this.X1 == undefined || this.X1 == 0;
    var np = system.planets.length;
    var x0 = +Infinity, x1 = 0;

    if (needX0 || needX1)
    {
	for (var i = 0; i < system.planets.length; ++i)
	{
	    if (this.isMoon(system.planets[i]))
	    {
		np--;
		continue;
	    }
	    if (system.planets[i].radius > x1)
		x1 = system.planets[i].radius;
	    if (system.planets[i].radius < x0)
		x0 = system.planets[i].radius;
	}
    }
    x[0] = needX0 ? x0 : this.X0;
    x[1] = needX1 ? x1 : this.X1;
    return np;
}

this.addPlanets = function(id)
{
    // Add planets in system dependent order. Id is used here in two
    // ways:
    // - Its lower bits determine the planets in planetinfo.plist
    // - It represents a permutation and hence the order how the selected
    //   planets are added (see wikipedia for factoradic) and Entity-IDs
    //   are assigned.
    var v = new Array(this.MaxPlanets), w = new Array(this.MaxPlanets);
    // Avoid huge factorials by grouping up to 12 planets.
    var mask = id;
    for (var i = 0; i < this.MaxPlanets && mask != 0 && id != 0; )
    {
	var m = 0, f = 1;
	// Type conversion prohibits use of the shift operator. Kind
	// of weird to apply "&" on a floating point number!
	for (; m < 12 && i < this.MaxPlanets; ++i, mask = Math.floor(mask / 2))
	    if ((mask & 1) != 0)
		v[m] = i, w[m] = m, f *= ++m;
	var fn = id % f;	// fn = sum(j in [0..m-1]:j!*c[j]), c[j] <= j
	id = Math.floor(id / f);
	for (; m > 0; --m)
	{
	    f /= m;
	    var c = Math.floor(fn / f);
	    fn -= f * c;
	    this.Log(this.Verbose,
		     "" + m + ":selecting oplanet" +
		     (v[w[c]] + 1) + "," + w[c] + "," +
		     c + "," + f + "," + fn);
	    system.addPlanet("oplanet" + (v[w[c]] + 1));
	    for (; c + 1 < m; ++c)
		w[c] = w[c + 1];
	}
    }
}

this.createOrbits = function()
{
    this.outerPlanets = 0;
    this.innerPlanets = 0;

    var s = system.sun;
    // Is there system.witchpoint?
    var w = system.shipsWithPrimaryRole("buoy-witchpoint")[0];
    var pm = system.mainPlanet;
    if (!s || !w || !pm)
	return;

    s.setPosition(pm.position.add(
	s.position.subtract(pm.position).multiply(this.FarsunMultiplier)));

    function PlanetVecs(s, w, p)
    {
	// "this" for sizzies:(
	this.s = s;
	this.sp = p.subtract(s);
	this.dist = this.sp.magnitude();
	// Rotation axis, perpendicular to the plane of triangle [s,p,w]
	this.rx = this.sp.cross(w.subtract(s)).direction();
	this.q = new Quaternion (1,0,0,0);
	this.rotateQ = function(q, axis, angle)
	{
	    // Ahruman's replacement for buggy Quaternion.rotate()
	    angle *= 0.5;
	    axis = axis.multiply(Math.sin(angle));
	    return q.multiply(Math.cos(angle), axis.x, axis.y, axis.z);
	}
	this.position = function(a, d)
	{
	    // Move sun to origin, rotate planet, and translate back again
	    return this.s.add(this.sp.rotateBy(
		this.rotateQ(this.q, this.rx, a)).multiply(d));
	}
    }
    var pv = new PlanetVecs(s.position, w.position, pm.position);

    function SystemId(name)
    {
	this.prime = 29;
	this.name = name;
	this.code = 0;

	this.charToInt = function(c)
	{
	    var l = "abcdefghijklmnopqrstuvwxyz";
	    var i = l.indexOf(c);
	    if (i >= 0)
		return i;
	    var u = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	    i = u.indexOf(c);
	    return i >= 0 ? i : 0; 
	}
	this.nextPrime = function()
	{
	    var p = this.prime;
	    if ((p & 1) == 0)
		++p;
	    for (;;)
	    {
		p += 2;
		for (var i = 3; ; i += 2)
		{
		    if (i * i > p)
			return this.prime = p;
		    if (p % i == 0)
			break;
		}
	    }
	}
	this.calcCode = function()
	{
	    // Use letters of n as digits of number with base prime.
	    var l = this.name.length;
	    var x = 0;
	    var pp = 1;
	    for (var i = 0; i < l; ++i)
	    {
		x += this.charToInt (this.name.charAt(i)) * pp;
		pp *= this.prime;
	    }
	    return this.code = x;
	}
    }
    // Derive numbers from system name
    var id = new SystemId(system.name);

    id.calcCode();
    this.Log (this.Verbose, system.name + ":id is " + id.code);
    this.addPlanets(id.code);
    this.innerPlanets = 1;

    var x = [+Infinity, 0];
    var np = this.planetsMinMax(x);
    if (np <= 1)
	return;

    function Map(x0, y0, x1, y1)
    {
	var m, X0 = x0, Y0 = y0;
	if (x0 == x1 || y0 == y1)
	{
	    m = (y0 + y1) / 2;
	    this.interpolate = function(x) { return m; }
	}
	else
	{
	    m = (y1 - y0) / (x1 - x0);
	    this.interpolate = function(x) { return Y0 + m * (x - X0); }
	}
    }
    var map = new Map(x[0], this.Y0, x[1], this.Y1);

    function PlanetWeights(p, s)
    {
	this.c = 0, this.p = p, this.s = s;
	this.product = function() { return this.s * this.c * this.p; }
	this.next = function() { this.p = this.c; }
    }
    var ow = new PlanetWeights(
	map.interpolate(pm.radius),
	pow(this.SpreadFactor, 1.0 / (np - 1)));
    var iw = new PlanetWeights(ow.p, ow.s);

    function InnerOrbitalPeriod(orbits, s)
    {
	this.t = orbits.Yr;
	var r = orbits.RandomOffsetFactor;
	r /= 1 + r;
	var f = orbits.Yr / (orbits.Y1 * orbits.Y1 * s);
	this.calc = function(id, iw)
	{
	    var mod = this.t * r;
	    return this.t - id.code % mod - f * iw.product();
	}
    }
    var iop = new InnerOrbitalPeriod(this, iw.s);

    function OuterOrbitalPeriod(orbits)
    {
	this.t = orbits.Yr;
	var r = orbits.RandomOffsetFactor;
	var f = orbits.Yr;
	this.calc = function(id, ow)
	{
	    var t = this.t + f * ow.product();
	    var mod = t * r;
	    if (mod != 0)
		t += id.code % mod;
	    return t;
	}
    }
    var oop = new OuterOrbitalPeriod(this);

    // Grrr, need an index sorted by Entity.ID as planets[] order may
    // change on different system entries.
    var idx = mkPlanetIndex();
    var today = clock.days + id.code;
    var a = 2.0 * Math.PI * (today % this.Yr) / this.Yr;
    var info = (180 * a / Math.PI).toFixed();

    this.Log(this.Verbose, system.name + "[" + "M/" + pm.ID + "]" + ":" +
	     today + "," +
	     ow.p.toFixed(4) + "," +
	     ow.s.toFixed(4) + "," +
	     (pm.radius  / 10).toFixed()+ "," +
	     (180 * a / Math.PI).toFixed());

    // Pretend to rotate the main planet.
    var pmOffset = pm.position.subtract(pv.position(a, 1));

    for (var k = 0; k < system.planets.length; ++k)
    {
	var i = idx[k];
	var pi = system.planets[i];
	if (pi.isMainPlanet || this.isMoon(pi))
	    continue;
	iw.c = ow.c = map.interpolate(pi.radius);

	id.nextPrime();
	id.calcCode();

	// Current planet's days per year. Try to make it an inner
	// planet. If that fails it will be an outer planet.
	var d;
	var t = iop.calc(id, iw);
	var inner = t > this.Yr / 3;
	if (inner)
	{
	    // Orbit Radius in planet-sun units (3. Kepler's law)
	    d = pow(t / this.Yr, 2.0 / 3.0);
	    inner =
		// idea: 0.3 * pv.dist / pm.r <= d * pv.dist / pi.r 
		pm.radius * d >= pi.radius * 0.3  &&
		d * pv.dist > 15 * s.radius;
	}

	if (!inner)
	{
	    t = oop.calc(id, ow);
	    d = pow(t / this.Yr, 2.0 / 3.0);
	}
	// Current angle in radians
	var a = 2.0 * Math.PI * (today % t) / t;
	var aa = (180.0 * a / Math.PI).toFixed();
	this.Log(this.Verbose,
		 system.name + "[" + i + "/" + pi.ID + "]" + ":" +
		 id.prime + "," +
		 id.code + "," +
		 (pi.radius / 10).toFixed() + "," +
		 ow.c.toFixed(4) + "," +
		 t.toFixed() + "," +
		 d.toFixed(4) + "," +
		 aa);
	pi.setPosition(pv.position(a, d).add(pmOffset));
	if (inner)
	{
	    iw.next(), iop.t = t, ++this.innerPlanets;
	    info = aa + "," + info;
	}
	else
	{
	    ow.next(), oop.t = t, ++this.outerPlanets;
	    info = info + "," + aa;
	}
    }
    s.setPosition(s.position.add(pmOffset));
    this.info = info;
}
